(*========================================================================*)
(*                                                                        *)
(*                 Compositional CompcertTSO                              *)
(*                                                                        *)
(*                 Chung-Kil Hur, MPI-SWS                                 *)
(*                 Viktor Vafeiadis, MPI-SWS                              *)
(*                 Derek Dreyer, MPI-SWS                                  *)
(*                                                                        *)
(*  Copyright 2010                                                        *)
(*                                                                        *)
(*  All rights reserved.  This file is distributed under the terms of     *)
(*  the INRIA Non-Commercial License Agreement.                           *)
(*                                                                        *)
(*========================================================================*)

Require Import Syntax.
Require Import List.
Require Import Relations.
Require Import Wellfounded.

Require Import Coqlib.
Require Import Libtactics.
Require Import MyTactics.
Require Import FiniteMap.

Require Import LangSpec.

Set Implicit Arguments.

Module WorldStructure (LBase : LangBaseSig).
  Module Export LSpec := LanguageSpecification LBase.

  Definition tidrel (A: Type) := tidtyp -> A -> A -> Prop.

  Inductive woth {wld:Type} (wtr: tidrel wld) : option tidtyp -> wld -> wld -> Prop :=
  | woth_refl : forall w otid, 
    woth(wtr) otid w w
  | woth_step : forall w w' w'' otid tid'
    (Hww': woth(wtr) otid w w') 
    (Htid': Some tid' <> otid) (Hw'w'': wtr tid' w' w''),
    woth(wtr) otid w w''.

  Record WorldStr (LS1 LS2 : LangSem) := mkWorldStr {
    wstates:> Type;
    wprv: tidrel wstates;
    wpub: tidrel wstates;
    winv: wstates -> LS1.(LS_mem) -> LS2.(LS_mem) -> Prop;

    wprv_preord: forall tid, preorder _ (wprv tid);
    wpub_preord: forall tid, preorder _ (wpub tid);
    wpub_min: forall tid, inclusion _ (woth wprv (Some tid)) (wpub tid);
    wpub_max: forall tid, inclusion _ (wpub tid) (woth wprv None)
  }.

  Record WorldStrG (LS1 LS2 : LangSem) := mkWorldStrG {
    wsg :> WorldStr LS1 LS2;
    
    wsg_funs: forall {w m1 m2}, 
      winv wsg w m1 m2 -> FMempty (LS_mfuns _ m1) /\ FMempty (LS_mfuns _ m2)
  }.

End WorldStructure.

(************************
   Program Equivalence
 ************************)

Module ProgramEquivalenceDef (LBase : LangBaseSig).

Module Export WStr := WorldStructure LBase.

(*****************
   Dereived steps
 *****************)

Inductive LS_taus LS tid : LS_state LS -> LS_state LS -> Prop :=
| LS_taus_refl: forall st, 
    LS_taus LS tid st st
| LS_taus_step: forall st st' st''
      (Htaus: LS_taus LS tid st st')
      (Hstep: LS_step LS tid Etau st' st''),
    LS_taus LS tid st st''.

Definition LS_tausteptau LS tid evt st st' : Prop :=
 exists sta stb, 
    LS_step LS tid evt sta stb /\
    LS_taus LS tid st sta /\
    LS_taus LS tid stb st'.

Definition LS_fails LS tid (evt: normal_event) st := 
  exists st' st'', 
    (LS_taus LS tid st st' \/ LS_tausteptau LS tid evt st st') /\
    LS_step LS tid Efail st' st''.

(***************************
   Language Specific World
 ***************************)

Record WfOrd := mkWfOrd {
  wfo_set :> Type;
  wfo_lt: wfo_set -> wfo_set -> Prop;
  wfo_transitive: transitive _ wfo_lt;
  wfo_wf: well_founded wfo_lt
}.

Record WorldStrL (LS1 LS2: LangSem) := mkWorldStrL {
  wsl :> WorldStr LS1 LS2;

  wsl_ord: WfOrd;

  wsl_efuns: forall {w m1 m2},
    winv wsl w m1 m2 -> FMequal (LS_mefuns _ m1) (LS_mefuns _ m2);
  wsl_funs: forall {w m1 m2},
    winv wsl w m1 m2 -> FMequal (LS_mfuns _ m1) (LS_mfuns _ m2)
}.

Lemma woth_preord : forall wld wtr otid, preorder _ (@woth wld wtr otid).
Proof. intros; split; vauto; red; induction 2; eauto using @woth_step. Qed.

(****************
     DModule
 ****************)

Structure DModM LS := mkDModM {
  dm_em : LS_mem LS;
  dm_lm : LS_mem LS;
  dm_m : LS_mem LS;
  
  dm_dunion: LS_mdunion _ dm_lm dm_em dm_m;
  dm_closed: LS_mclosed _ dm_m
}.

Structure DMod LS := mkDMod {
  dmts : LS_tstate LS;
  dmod :> DModM LS
}.

Structure DModO LS := mkDModO {
  dmtso : option (LS_tstate LS);
  dmodo :> DModM LS
}.

Definition dmodo_cast LS (dm: DMod LS) : DModO LS :=
  mkDModO (Some dm.(dmts)) dm.

Coercion dmodo_cast : DMod >-> DModO.

Definition dmod_ts LS tid (dm: DModO LS) st :=
  FMfind (LS_st_tss LS st) tid = dm.(dmtso).

Definition dmod_mem LS (dm: DModM LS) st :=
  LS_st_mem LS st = dm.(dm_m).

Definition dmod_st LS tid (dm: DModO LS) st :=
  dmod_ts tid dm st /\ dmod_mem dm st.

Definition retn LS tid rty rtc v (dm: DMod LS) :=
  LS_funret LS tid rty v rtc dm.(dm_m) dm.(dmts).

Definition call LS tid f sig args rtc (dm: DMod LS) :=
  FMfind (LS_mefuns _ dm.(dm_em)) f = Some sig /\
  LS_funapp LS tid f sig args rtc dm.(dm_m) dm.(dmts).

(********************** 
      Simulation
 **********************)

Section Simulation.

  Variables LS1 LS2 : LangSem.
  Variable W: WorldStrL LS1 LS2.

  Record RetContR := mkRetContR {
    rc_rtyp : option gtyp;
    rc_rtc1 : LS_tcont LS1;
    rc_rtc2 : LS_tcont LS2
  }.

  Record RetCont := mkRetCont {
    rc_retw : W;
    rc_cont :> RetContR
  }.

  Definition tsrel :=
    tidtyp -> option RetCont ->
    W -> wsl_ord W -> LS_tstate LS1 -> LS_tstate LS2 -> Prop.

  Definition tcrel :=
    tidtyp -> option RetCont -> RetCont -> Prop.

  Definition tsrel_le (R1 R2: tsrel) : Prop :=
    forall tid ret w ord ts1 ts2 
      (HR1: R1 tid ret w ord ts1 ts2),
      R2 tid ret w ord ts1 ts2.

  Definition tcrel_le (R1 R2: tcrel) : Prop :=
    forall tid ret rc 
      (HRc1: R1 tid ret rc),
      R2 tid ret rc.

  Definition memext_prv (em1: LS_mem LS1) (em2: LS_mem LS2) em1' em2' :=
    em1 = em1' /\ em2 = em2'.

  Definition memext_r em1 em2 :=
    LS_mem_external LS1 em1 /\ 
    LS_mem_external LS2 em2 /\
    FMequal (LS_mefuns _ em1) (LS_mefuns _ em2) /\
    FMequal (LS_mfuns _ em1) (LS_mfuns _ em2).
    
  Definition winv_mem (w: W) (dm1: DModM LS1) (dm2: DModM LS2) : Prop :=
    winv _ w dm1.(dm_lm) dm2.(dm_lm) /\ memext_r dm1.(dm_em) dm2.(dm_em).

  Definition dmod_st_r tid dm1 dm2 st1 st2 :=
    @dmod_st LS1 tid dm1 st1 /\ @dmod_st LS2 tid dm2 st2.

  Definition retn_r tid ret v w dm1 dm2 :=
    option_default False (fun r =>
      wpub _ tid (rc_retw r) w /\
      @retn _ tid (rc_rtyp r) (rc_rtc1 r) v dm1 /\
      @retn _ tid (rc_rtyp r) (rc_rtc2 r) v dm2) ret. 

  Definition call_r tid f sig args rtc1 rtc2 dm1 dm2 :=
    @call LS1 tid f sig args rtc1 dm1 /\
    @call LS2 tid f sig args rtc2 dm2.

  Definition ret_external (ret: option RetCont) :=
    option_default True (fun r : RetCont => 
      LS_tc_external _ (rc_rtc1 r) /\
      LS_tc_external _ (rc_rtc2 r)
    ) ret.

  Definition ts_matched st1 st2 : Prop := 
    forall tid, 
      FMfind (LS_st_tss LS1 st1) tid <> None <-> FMfind (LS_st_tss LS2 st2) tid <> None.

  Definition step_type := 
    tidtyp -> event -> LS_state LS2 -> LS_state LS2 -> Prop.
  
  Definition equal_to_ret : tcrel := fun tid ret r =>
    option_default False (fun r' =>
      wpub _ tid (rc_retw r') (rc_retw r) /\ 
      rc_rtyp r = rc_rtyp r' /\ rc_rtc1 r = rc_rtc1 r' /\ rc_rtc2 r = rc_rtc2 r') ret.

  Definition ret_or_Rc (Rc: tcrel) : tcrel := fun tid ret r =>
    equal_to_ret tid ret r \/ Rc tid ret r.

  Definition dmR_keep (R: tsrel) tid ret w' ord stepped1 stepped2 dm1' dm2'
    :=
    exists ord',
      (stepped1 = false ->
        (stepped2 = true -> W.(wsl_ord).(wfo_lt) ord' ord) /\
        (stepped2 = false -> ord' = ord)) /\
      R tid ret w' ord' dm1'.(dmts) dm2'.(dmts).
 
  Definition dmR_retn tid ret w' stepped1 dm1' dm2' v :=
    stepped1 = true /\ retn_r tid ret v w' dm1' dm2'. 

  Definition dmR_call Rc tid ret w' stepped1 dm1' dm2'
    f sig args rtc1 rtc2 
    :=
    stepped1 = true /\ 
    call_r tid f sig args rtc1 rtc2 dm1' dm2' /\
    ret_or_Rc Rc tid ret (mkRetCont w' (mkRetContR sig.(gsig_res) rtc1 rtc2)).

  Definition dmR (R:tsrel) (Rc: tcrel) tid ret w' ord stepped1 stepped2 
    (dmo1': DModO _) (dmo2': DModO _)
    :=
    winv_mem w' dmo1' dmo2' /\ 
    forall (dm2': DMod _), dmo2' = dm2' ->
    exists dm1': DMod _, dmo1' = dm1' /\
     (dmR_keep R tid ret w' ord stepped1 stepped2 dm1' dm2' \/
      (exists v, 
        dmR_retn tid ret w' stepped1 dm1' dm2' v) \/ 
      (exists f sig args rtc1 rtc2, 
        dmR_call Rc tid ret w' stepped1 dm1' dm2' f sig args rtc1 rtc2)).

  Definition tsR_postcond_success R Rc tid ret w w' ord stepped1 stepped2
    (dm1: DMod _) (dm2: DMod _) (dmo1': DModO _) (dmo2': DModO _) st1 st1' st2' evt
    :=
    ((stepped1 = true -> LS_tausteptau _ tid evt st1 st1') /\
     (stepped1 = false -> st1' = st1 /\ evt = Etau)) /\
    @dmod_st_r tid dmo1' dmo2' st1' st2' /\
    wprv _ tid w w' /\ 
    memext_prv dm1.(dm_em) dm2.(dm_em) dmo1'.(dm_em) dmo2'.(dm_em) /\
    dmR R Rc tid ret w' ord stepped1 stepped2 dmo1' dmo2' /\
    (forall tidnew, evt = Etstart tidnew -> 
      exists tsnew1 tsnew2 ordnew,
      FMfind (LS_st_tss _ st1') tidnew = Some tsnew1 /\
      FMfind (LS_st_tss _ st2') tidnew = Some tsnew2 /\
      R tidnew None w' ordnew tsnew1 tsnew2).

  Definition tsR_postcond R Rc tid (evt: event) ret w ord stepped2 
    (dm1: DMod _) (dm2: DMod _) st1 st2' 
    :=
    (LS_fails LS1 tid (e2ne evt) st1)
    \/
    (evt <> Efail /\ 
     exists st1' w' stepped1 dmo1' dmo2',
     tsR_postcond_success R Rc tid ret w w' ord stepped1 stepped2
       dm1 dm2 dmo1' dmo2' st1 st1' st2' evt).

  Definition tsR_simulated (sem_step: step_type) R Rc tid ret w ord 
    (dm1: DMod _) (dm2: DMod _)
    :=
    forall evt st1 st2 st2'
      (Htsmat: ts_matched st1 st2)
      (Hst: dmod_st_r tid dm1 dm2 st1 st2)
      (Hstep: sem_step tid evt st2 st2')
      (Hnotoom: evt <> Eoutofmemory),
    tsR_postcond R Rc tid evt ret w ord true dm1 dm2 st1 st2'.

  Definition tsR_meta (sem_step: step_type) R Rc : tsrel 
    := fun tid ret wcrr ord ts1 ts2  =>

    LS_ts_normal LS1 ts1 /\ LS_ts_normal LS2 ts2 
    /\
    forall w (dm1: DMod _) (dm2: DMod _)
      (Heq1: dm1.(dmts) = ts1) 
      (Heq2: dm2.(dmts) = ts2)
      (Hret: ret_external ret)      
      (Hothw: woth(wprv _) (Some tid) wcrr w)
      (Hinvw: winv_mem w dm1 dm2),

    tsR_simulated sem_step R Rc tid ret w ord dm1 dm2.

  Definition tsR := tsR_meta (LS_step _).

  Definition tcR (R: tsrel) : tcrel := fun tid ret r =>

    LS_tc_internal _ (rc_rtc1 r) /\ LS_tc_internal _ (rc_rtc2 r) 
    /\
    forall v w (dm1: DMod _) (dm2: DMod _)
      (Hinvw : winv_mem w dm1 dm2)
      (Hretn : retn_r tid (Some r) v w dm1 dm2),
    exists ord,
    R tid ret w ord dm1.(dmts) dm2.(dmts).

  Definition tsR_woth_closed (R: tsrel) : Prop := 
    forall tid ret w ord ts1 ts2 w',
      R tid ret w ord ts1 ts2 ->
      woth (wprv _) (Some tid) w w' ->
    R tid ret w' ord ts1 ts2.

  Definition tcR_wpub_closed (Rc: tcrel) : Prop := 
    forall tid ret w w' rc,
      Rc tid ret (mkRetCont w rc) ->
      wpub _ tid w w' ->
    Rc tid ret (mkRetCont w' rc).

  Definition simulated (R: tsrel) (Rc: tcrel) : Prop := 
    tsrel_le R (tsR R Rc) /\ tcrel_le Rc (tcR R) /\
    tsR_woth_closed R /\ tcR_wpub_closed Rc.

End Simulation.
Implicit Arguments memext_r [LS1 LS2].
Implicit Arguments mkRetCont [LS1 LS2].
Implicit Arguments mkRetContR [LS1 LS2].
Implicit Arguments ts_matched [LS1 LS2].

End ProgramEquivalenceDef.

